use std::{
    ffi::OsString,
    net::IpAddr,
    ops::{Deref, DerefMut},
    path::PathBuf,
    time::Duration,
};

use derive_more::Display;
use osstr_traits_derive::OsDisplay;
use serde::{Deserialize, Serialize};

use crate::collection::{CommandArgs, DelimitedVec};

type OsStrDisplay<'a> = std::ffi::os_str::Display<'a>;
type PathDisplay<'a> = std::path::Display<'a>;

#[derive(Clone, Copy, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum ConsistencyCheckSourceMode {
    #[display("local")]
    Local,
    #[display("cacheserver")]
    CacheServer,
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("{0} {1}", DelimitedVec::<PathDisplay>::from(exports), path.display())]
#[os_display("{exports} {path}", exports = DelimitedVec::<&PathBuf, ' '>::from_iter(exports), path = path)]
pub struct ExportPackage {
    path: PathBuf,
    exports: Vec<PathBuf>,
}

#[derive(Clone, Copy, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum PlatformTextureFormat {
    #[display("dxt")]
    Dxt,
    #[display("pvrtc")]
    #[deprecated(note = "PVRTC format is deprecated. Use ASTC or ETC format instead.")]
    Pvrtc,
    #[display("atc")]
    Atc,
    #[display("etc")]
    Etc,
    #[display("etc2")]
    Etc2,
    #[display("astc")]
    Astc,
}

#[derive(Clone, Copy, Debug, Default, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum TextureCompression {
    #[default]
    NoOverride,
    ForceUncompressed,
    ForceFastCompressor,
    ForceNoCrunchCompression,
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum SharedLogLevel {
    #[display("debug")]
    Debug,
    #[display("info")]
    Info,
    #[display("notice")]
    Notice,
    #[display("fatal")]
    Fatal,
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
pub enum PerforceArgument {
    #[display("vcPerforceHost {0}", _0.display())]
    #[os_display("vcPerforceHost {_0}")]
    Host(OsString),
    #[display("vcSharedLogLevel {_0}")]
    #[os_display("vcSharedLogLevel {_0}")]
    SharedLogLevel(SharedLogLevel),
    #[display("vcPerforceServer {0}", _0.display())]
    #[os_display("vcPerforceServer {_0}")]
    Server(OsString),
    #[display("vcPerforceWorkspace {0}", _0.display())]
    #[os_display("vcPerforceWorkspace {_0}")]
    Workspace(OsString),
    #[display("vcPerforceUsername {0}", _0.display())]
    #[os_display("vcPerforceUsername {_0}")]
    Username(OsString),
    #[display("vcPerforcePassword {0}", _0.display())]
    #[os_display("vcPerforcePassword {_0}")]
    Password(OsString),
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum VcsMode {
    #[display("\"Visible Meta Files\"")]
    VisibleMetaFiles,
    #[display("\"Hidden Meta Files\"")]
    HiddenMetaFiles,
    #[display("Perforce {_0}")]
    #[os_display("Perforce {_0}")]
    Perforce(CommandArgs<PerforceArgument>),
    #[display("PlasticSCM")]
    PlasticScm,
    #[display("{0}", _0.display())]
    #[os_display("{_0}")]
    Other(OsString),
}

#[derive(Debug, Display, Clone, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum ConfigurationArgument {
    #[display("createProject \"{0}\"", _0.display())]
    #[os_display("-createProject \"{_0}\"")]
    CreateProject(PathBuf),
    #[display("consistencyCheck")]
    ConsistencyCheck,
    #[display("consistencyCheckSourceMode {_0}")]
    ConsistencyCheckSourceMode(ConsistencyCheckSourceMode),
    #[display("disable-assembly-updater {0}", DelimitedVec::<PathDisplay, ' '>::from(_0))]
    #[os_display("-disable-assembly-updater {}", DelimitedVec::<&PathBuf, ' '>::from_iter(_0))]
    DisableAssemblyUpdater(Vec<PathBuf>),
    #[display("disable-gpu-skinning")]
    DisableGpuSkinning,
    #[display("disable-playback-engines {0}", DelimitedVec::<OsStrDisplay, ' '>::from(_0))]
    #[os_display("-disable-playback-engines {}", DelimitedVec::<&std::ffi::OsString, ' '>::from_iter(_0))]
    DisablePlaybackEngines(Vec<OsString>),
    #[display("executeMethod {0}", _0.display())]
    #[os_display("-executeMethod {_0}")]
    ExecMethod(OsString),
    #[display("exportPackage {_0}")]
    #[os_display("-exportPackage {_0}")]
    ExportPackage(ExportPackage),
    #[display("importPackage {0}", _0.display())]
    #[os_display("-importPackage {_0}")]
    ImportPackage(PathBuf),
    #[display("job-worker-count {_0}")]
    JobWorkerCount(usize),
    #[display("gc-helper-count {_0}")]
    GcHelperCount(usize),
    #[display("logFile {0}", _0.display())]
    #[os_display("-logFile {_0}")]
    LogFile(PathBuf),
    #[display("noUpm")]
    NoUpm,
    #[display("openfile {0}", _0.display())]
    #[os_display("-openfile {_0}")]
    OpenFile(PathBuf),
    #[display("password {0}", _0.display())]
    #[os_display("-password {_0}")]
    Password(OsString),
    #[display("projectPath {0}", _0.display())]
    #[os_display("-projectPath {_0}")]
    ProjectPath(PathBuf),
    #[display("quit")]
    Quit,
    #[display("releaseCodeOptimization")]
    ReleaseCodeOptimization,
    #[display("setDefaultPlatformTextureFormat {_0}")]
    SetDefaultPlatformTextureFormat(PlatformTextureFormat),
    #[display("overrideMaxTextureSize {_0}")]
    OverrideMaxTextureSize(usize),
    #[display("overrideTextureCompression {_0}")]
    OverrideTextureCompression(TextureCompression),
    #[display("silent-crashes")]
    SilentCrashes,
    #[display("upmLogFile {0}", _0.display())]
    #[os_display("-upmLogFile {_0}")]
    UpmLogFile(PathBuf),
    #[display("username {0}", _0.display())]
    #[os_display("-username {_0}")]
    Username(OsString),
    #[display("vcsMode {_0}")]
    #[os_display("-vcsMode {_0}")]
    VcsMode(VcsMode),
    #[display("version")]
    Version,
    #[display("timestamps")]
    Timestamps,
}

#[derive(Debug, Clone, Copy, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum BatchModeArgument {
    #[display("accent-apiupdate")]
    AcceptApiUpdate,
    #[display("batchmode")]
    BatchMode,
    #[display("ignorecompilererrors")]
    IgnoreCompilerErrors,
    #[display("nographics")]
    NoGraphics,
}

#[derive(Debug, Clone, Copy, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum BuildTarget {
    #[display("win64")]
    Win64,
    #[display("win")]
    Win,
    #[display("osxuniversal")]
    OsxUniversal,
    #[display("linux64")]
    Linux64,
    #[display("android")]
    Android,
    #[display("ios")]
    IOs,
    #[display("webgl")]
    WebGl,
    #[display("tvos")]
    TvOs,
    #[display("windowsstoreapps")]
    WindowsStoreApps,
    #[display("cloudrendering")]
    CloudRendering,
    #[display("visionos")]
    VisionOs,
}

#[derive(Debug, Clone, Copy, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum BuildSubtarget {
    Player,
    Server,
}

#[derive(Debug, Display, Clone, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
pub enum BuildArgument {
    #[display("activeBuildProfile \"{0}\"", _0.display())]
    #[os_display("-activeBuildProfile \"{_0}\"")]
    ActiveBuildProfile(PathBuf),
    #[display("build \"{0}\"", _0.display())]
    #[os_display("-build \"{_0}\"")]
    Build(PathBuf),
    #[display("buildLinux64Player \"{0}\"", _0.display())]
    #[os_display("-buildLinux64Player \"{_0}\"")]
    BuildLinux64Player(PathBuf),
    #[display("buildLinuxHeadlessSimulation \"{0}\"", _0.display())]
    #[os_display("-buildLinuxHeadlessSimulation \"{_0}\"")]
    BuildLinuxHeadlessSimulation(PathBuf),
    #[display("buildOSXUniversalPlayer \"{0}\"", _0.display())]
    #[os_display("-buildOSXUniversalPlayer \"{_0}\"")]
    BuildOsxUniversalPlayer(PathBuf),
    #[display("buildTarget \"{_0}\"")]
    BuildTarget(BuildTarget),
    #[display("standaloneBuildSubtarget \"{_0}\"")]
    #[os_display("-standaloneBuildSubtarget \"{_0}\"")]
    StandaloneBuildSubtarget(BuildSubtarget),
    #[display("buildWindowsPlayer \"{0}\"", _0.display())]
    #[os_display("-buildWindowsPlayer \"{_0}\"")]
    BuildWindowsPlayer(PathBuf),
    #[display("buildWindows64Player \"{0}\"", _0.display())]
    #[os_display("-buildWindows64Player \"{_0}\"")]
    BuildWindows64Player(PathBuf),
}
#[derive(Debug, Display, Clone, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum CacheServerArgument {
    #[display("EnableCacheServer")]
    Enable,
    #[display("cacheServerEndpoint {_0}:{_1}")]
    Endpoint(IpAddr, u16),
    #[display("cacheServerNamespacePrefix \"{0}\"", _0.display())]
    #[os_display("-cacheServerNamespacePrefix \"{_0}\"")]
    NamespacePrefix(OsString),
    #[display("cacheServerEnableDownload {_0}")]
    EnableDownload(bool),
    #[display("cacheServerEnableUpload {_0}")]
    EnableUpload(bool),
    #[display("cacheServerWaitForConnection {0}", _0.as_millis())]
    WaitForConnection(Duration),
    #[display("cacheServerWaitForUploadCompletion")]
    WaitForUploadCompletion,
    #[display("cacheServerDownloadBatchSize {_0}")]
    DownloadBatchSize(usize),
    #[display("cacheServerUploadAllRevisions")]
    UploadAllRevisions,
    #[display("cacheServerUploadExistingShaderCache")]
    UploadExistingShaderCache,
}

#[derive(Debug, Default, Display, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[os_display(from_display)]
pub enum StackTraceLogType {
    None,
    #[display("\"Script Only\"")]
    ScriptOnly,
    #[default]
    Full,
}

#[derive(Debug, Display, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum DebugArgument {
    #[display("disableManagedDebugger")]
    DisableManagedDebugger,
    #[display("diag-debug-shader-compiler")]
    DiagDebugShaderCompiler,
    #[display("debugCodeOptimization")]
    DebugCodeOptimization,
    #[display("enableCodeCoverage")]
    EnableCodeCoverage,
    #[display("force-d3d12-debug")]
    ForceD3d12Debug,
    #[display("force-d3d12-debug-gbv")]
    ForceD3d12DebugGbv,
    #[display("force-vulkan-layers")]
    ForceVulkanLayers,
    #[display("StackTraceLogType {_0}")]
    #[os_display("-StackTraceLogType {_0}")]
    StackTraceLogType(StackTraceLogType),
    #[display("log-memory-performance-stats")]
    LogMemoryPerformanceStats,
    #[display("wait-for-managed-debugger")]
    WaitForManagedDebugger,
    #[display("wait-for-native-debugger")]
    WaitForNativeDebugger,
}

#[derive(Clone, Copy, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum GraphicsApiArgument {
    #[display("force-clamped")]
    ForceClamped,
    #[display("force-d3d11")]
    ForceD3d11,
    #[display("force-d3d12")]
    ForceD3d12,
    #[display("force-device-index")]
    ForceDeviceIndex,
    #[display("force-glcore")]
    ForceGlCore,
    #[display("force-glcoreXY")]
    ForceGlCoreXy,
    #[display("force-gles")]
    ForceGlEs,
    #[display("force-glesXY")]
    ForceGlEsXy,
    #[display("force-opengl")]
    ForceOpenGl,
    #[display("force-vulkan")]
    ForceVulkan,
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum LicenseArgument {
    #[display("createManualActivationFile")]
    CreateManualActivationFile,
    #[display("manualLicenseFile {0}", _0.display())]
    #[os_display("-manualLicenseFile {_0}")]
    ManualLicenseFile(PathBuf),
    #[display("returnlicense")]
    ReturnLicense,
    #[display("serial {0}", _0.display())]
    #[os_display("-serial {_0}")]
    Serial(OsString),
}

#[derive(Clone, Copy, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum MetalArgument {
    #[display("force-low-power-device")]
    ForceLowPowerDevice,
    #[display("force-metal")]
    ForceMetal,
    #[display("enable-metal-capture")]
    EnableMetalCapture,
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum ProfilerArgument {
    #[display("deepprofiling")]
    DeepProfiling,
    #[display("profiler-enable")]
    Enable,
    #[display("profiler-log-file {0}", _0.display())]
    #[os_display("-profiler-log-file {_0}")]
    LogFile(PathBuf),
    #[display("profiler-capture-frame-count {_0}")]
    CaptureFrameCount(usize),
    #[display("profiler-maxusedmemory {_0}")]
    MaxUsedMemory(usize),
}

#[derive(Clone, Debug, Display, PartialEq, Eq, PartialOrd, Ord, OsDisplay)]
#[display("-{_variant}")]
#[os_display(from_display)]
pub enum EditorSpecialArgument {
    #[display("enableIncompatibleAssetDowngrade")]
    EnableIncompatibleAssetDowngrade,
    #[display("giCustomCacheLocation {0}", _0.display())]
    #[os_display("-giCustomCacheLocation {_0}")]
    GiCustomCacheLocation(PathBuf),
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum UnityArgument {
    Configuration(ConfigurationArgument),
    BatchMode(BatchModeArgument),
    Build(BuildArgument),
    CacheServer(CacheServerArgument),
    Debug(DebugArgument),
    GraphicsApi(GraphicsApiArgument),
    License(LicenseArgument),
    Metal(MetalArgument),
    Profiler(ProfilerArgument),
    Editor(EditorSpecialArgument),
    Extra(OsString),
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct UnityArguments(Vec<UnityArgument>);

impl Deref for UnityArguments {
    type Target = Vec<UnityArgument>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl DerefMut for UnityArguments {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl Default for UnityArguments {
    fn default() -> Self {
        Self(Vec::new())
    }
}

macro_rules! unity_args {
    () => {
        UnityArguments::default()
    };

    ($arg:expr $(, $rest:expr)* $(,)?) => {
        {
            let mut args = Vec::new();
            $(
                #[allow(unreachable_patterns)]
                match $rest {
                    arg @ ConfigArguments::CreateProject(_) |
                    arg @ ConfigArguments::ConsistencyCheck |
                    arg @ ConfigArguments::ConsistencyCheckSourceMode(_) |
                    arg @ ConfigArguments::DisableAssemblyUpdater(_) |
                    arg @ ConfigArguments::DisableGpuSkinning |
                    arg @ ConfigArguments::DisablePlaybackEngines(_) |
                    arg @ ConfigArguments::ExecMethod(_) |
                    arg @ ConfigArguments::ExportPackage(_) |
                    arg @ ConfigArguments::ImportPackage(_) |
                    arg @ ConfigArguments::JobWorkerCount(_) |
                    arg @ ConfigArguments::GcHelperCount(_) |
                    arg @ ConfigArguments::LogFile(_) |
                    arg @ ConfigArguments::NoUpm |
                    arg @ ConfigArguments::OpenFile(_) |
                    arg @ ConfigArguments::Password(_) |
                    arg @ ConfigArguments::ProjectPath(_) |
                    arg @ ConfigArguments::Quit |
                    arg @ ConfigArguments::ReleaseCodeOptimization |
                    arg @ ConfigArguments::SetDefaultPlatformTextureFormat |
                    arg @ ConfigArguments::OverrideMaxTextureSize(_) |
                    arg @ ConfigArguments::OverrideTextureCompression(_) |
                    arg @ ConfigArguments::SilentCrashes |
                    arg @ ConfigArguments::UpmLogFile(_) |
                    arg @ ConfigArguments::Username(_) |
                    arg @ ConfigArguments::VcsMode |
                    arg @ ConfigArguments::Version |
                    arg @ ConfigArguments::Timestamps => {
                        args.push(UnityArgument::Configuration(arg));
                    },

                    arg @ BatchModeArguments::AcceptApiUpdate |
                    arg @ BatchModeArguments::BatchMode |
                    arg @ BatchModeArguments::IgnoreCompileErrors |
                    arg @ BatchModeArguments::NoGraphics => {
                        args.push(UnityArgument::BatchMode(arg));
                    },

                    arg @ BuildArguments::ActiveBuildProfile(_) |
                    arg @ BuildArguments::Build(_) |
                    arg @ BuildArguments::BuildLinux64Player(_) |
                    arg @ BuildArguments::BuildLinuxHeadlessSimulation(_) |
                    arg @ BuildArguments::BuildOSXUniversalPlayer(_) |
                    arg @ BuildArguments::BuildTarget(_) |
                    arg @ BuildArguments::StandaloneBuildSubtarget(_) |
                    arg @ BuildArguments::BuildWindowsPlayer(_) |
                    arg @ BuildArguments::BuildWindows64Player(_) => {
                        args.push(UnityArgument::Build(arg));
                    },

                    arg @ CacheServerArguments::Enable |
                    arg @ CacheServerArguments::Endpoint(_, _) |
                    arg @ CacheServerArguments::NamespacePrefix(_) |
                    arg @ CacheServerArguments::EnableDownload(_) |
                    arg @ CacheServerArguments::EnableUpload(_) |
                    arg @ CacheServerArguments::WaitForConnection(_) |
                    arg @ CacheServerArguments::WaitForUploadCompletion |
                    arg @ CacheServerArguments::DownloadBatchSize(_) |
                    arg @ CacheServerArguments::UploadAllRevisions |
                    arg @ CacheServerArguments::UploadExistingShaderCache => {
                        args.push(UnityArgument::CacheServer(arg));
                    },

                    arg @ DebugArguments::DisableManagedDebugger |
                    arg @ DebugArguments::DiagDebugShaderCompiler |
                    arg @ DebugArguments::DebugCodeOptimization |
                    arg @ DebugArguments::EnableCodeCoverage |
                    arg @ DebugArguments::ForceD3d12Debug |
                    arg @ DebugArguments::ForceD3d12DebugGbv |
                    arg @ DebugArguments::ForceVulkanLayers |
                    arg @ DebugArguments::StackTraceLogType(_) |
                    arg @ DebugArguments::LogMemoryPerformanceStats |
                    arg @ DebugArguments::WaitForManagedDebugger |
                    arg @ DebugArguments::WaitForNativeDebugger => {
                        args.push(UnityArgument::Debug(arg));
                    },

                    arg @ GraphicsApiArguments::ForceClamped |
                    arg @ GraphicsApiArguments::ForceD3d11 |
                    arg @ GraphicsApiArguments::ForceD3d12 |
                    arg @ GraphicsApiArguments::ForceDeviceIndex |
                    arg @ GraphicsApiArguments::ForceGlCore |
                    arg @ GraphicsApiArguments::ForceGlCoreXy |
                    arg @ GraphicsApiArguments::ForceGlEs |
                    arg @ GraphicsApiArguments::ForceGlEsXy |
                    arg @ GraphicsApiArguments::ForceOpenGl |
                    arg @ GraphicsApiArguments::ForceVulkan => {
                        args.push(UnityArgument::GraphicsApi(arg));
                    },

                    arg @ LicenseArguments::CreateManualActivationFile |
                    arg @ LicenseArguments::ManualLicenseFile(_) |
                    arg @ LicenseArguments::ReturnLicense |
                    arg @ LicenseArguments::Serial(_) => {
                        args.push(UnityArgument::License(arg));
                    },

                    arg @ MetalArguments::ForceLowPowerDevice |
                    arg @ MetalArguments::ForceMetal |
                    arg @ MetalArguments::EnableMetalCapture => {
                        args.push(UnityArgument::Metal(arg));
                    },

                    arg @ ProfilerArguments::DeepProfiling |
                    arg @ ProfilerArguments::Enable |
                    arg @ ProfilerArguments::LogFile(_) |
                    arg @ ProfilerArguments::CaptureFrameCount(_) |
                    arg @ ProfilerArguments::MaxUsedMemory(_) => {
                        args.push(UnityArgument::Profiler(arg));
                    },

                    arg @ EditorSpecialArguments::EnableIncompatibleAssetDowngrade |
                    arg @ EditorSpecialArguments::GiCustomCacheLocation(_) => {
                        args.push(UnityArgument::Editor(arg));
                    },

                    arg @ UnityArgument::Configuration(_) |
                    arg @ UnityArgument::BatchMode(_) |
                    arg @ UnityArgument::Build(_) |
                    arg @ UnityArgument::CacheServer(_) |
                    arg @ UnityArgument::Debug(_) |
                    arg @ UnityArgument::GraphicsApi(_) |
                    arg @ UnityArgument::License(_) |
                    arg @ UnityArgument::Metal(_) |
                    arg @ UnityArgument::Profiler(_) |
                    arg @ UnityArgument::Editor(_) |
                    arg @ UnityArgument::Extra(_) => {
                        args.push(arg);
                    },

                    s: OsString => {
                        args.push(UnityArgument::Extra(s));
                    },

                    _ => compile_error!("Unsupported argument type in unity_args! macro. Ensure it's a valid argument enum variant or a String."),
                }
            )*

            UnityArguments(args)
        }
    };
}

#[derive(Debug, Serialize, Deserialize)]
pub struct EditorCommand {}

#[cfg(test)]
mod tests {
    use std::{ffi::OsString, path::PathBuf};

    use osstr_traits::ToOsString;

    use crate::{
        collection::CommandArgs,
        editor_command::{
            ConfigurationArgument, ExportPackage, PerforceArgument, PlatformTextureFormat,
            SharedLogLevel, TextureCompression, VcsMode,
        },
    };

    #[test]
    fn config_args() {
        let arg = ConfigurationArgument::CreateProject(PathBuf::from("./test-project"));
        assert_eq!(arg.to_os_string(), "-createProject \"./test-project\"");

        let arg = ConfigurationArgument::ConsistencyCheck;
        assert_eq!(arg.to_os_string(), "-consistencyCheck");

        let arg = ConfigurationArgument::ConsistencyCheckSourceMode(
            super::ConsistencyCheckSourceMode::Local,
        );
        assert_eq!(arg.to_os_string(), "-consistencyCheckSourceMode local");

        let arg = ConfigurationArgument::ConsistencyCheckSourceMode(
            super::ConsistencyCheckSourceMode::CacheServer,
        );
        assert_eq!(
            arg.to_os_string(),
            "-consistencyCheckSourceMode cacheserver"
        );

        let arg =
            ConfigurationArgument::DisableAssemblyUpdater(vec!["a.dll".into(), "b.dll".into()]);
        assert_eq!(arg.to_os_string(), "-disable-assembly-updater a.dll b.dll");

        let arg = ConfigurationArgument::DisableGpuSkinning;
        assert_eq!(arg.to_os_string(), "-disable-gpu-skinning");

        let arg = ConfigurationArgument::DisablePlaybackEngines(vec![
            "AndroidPlayer".into(),
            "iOSSupport".into(),
        ]);
        assert_eq!(
            arg.to_os_string(),
            "-disable-playback-engines AndroidPlayer iOSSupport"
        );

        let arg = ConfigurationArgument::ExecMethod("NamespaceName.ClassName.MethodName".into());
        assert_eq!(
            arg.to_os_string(),
            "-executeMethod NamespaceName.ClassName.MethodName"
        );

        let arg = ConfigurationArgument::ExportPackage(ExportPackage {
            path: "export.unitypackage".into(),
            exports: vec!["Assets/a.asset".into(), "Assets/b.asset".into()],
        });
        assert_eq!(
            arg.to_os_string(),
            "-exportPackage Assets/a.asset Assets/b.asset export.unitypackage"
        );

        let arg = ConfigurationArgument::ImportPackage(PathBuf::from("./package.unitypackage"));
        assert_eq!(arg.to_os_string(), "-importPackage ./package.unitypackage");

        let arg = ConfigurationArgument::JobWorkerCount(8usize);
        assert_eq!(arg.to_os_string(), "-job-worker-count 8");

        let arg = ConfigurationArgument::GcHelperCount(12usize);
        assert_eq!(arg.to_os_string(), "-gc-helper-count 12");

        let arg = ConfigurationArgument::LogFile(PathBuf::from("/var/log/unity/editor.log"));
        assert_eq!(arg.to_os_string(), "-logFile /var/log/unity/editor.log");

        let arg = ConfigurationArgument::NoUpm;
        assert_eq!(arg.to_os_string(), "-noUpm");

        let arg = ConfigurationArgument::OpenFile(PathBuf::from("./scene.unity"));
        assert_eq!(arg.to_os_string(), "-openfile ./scene.unity");

        let arg = ConfigurationArgument::Password(OsString::from("hunter12"));
        assert_eq!(arg.to_os_string(), "-password hunter12");

        let arg = ConfigurationArgument::ProjectPath(PathBuf::from("./my-project/"));
        assert_eq!(arg.to_os_string(), "-projectPath ./my-project/");

        let arg = ConfigurationArgument::Quit;
        assert_eq!(arg.to_os_string(), "-quit");

        let arg = ConfigurationArgument::ReleaseCodeOptimization;
        assert_eq!(arg.to_os_string(), "-releaseCodeOptimization");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Dxt);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat dxt");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Atc);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat atc");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Etc);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat etc");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Etc2);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat etc2");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Pvrtc);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat pvrtc");

        let arg =
            ConfigurationArgument::SetDefaultPlatformTextureFormat(PlatformTextureFormat::Astc);
        assert_eq!(arg.to_os_string(), "-setDefaultPlatformTextureFormat astc");

        let arg = ConfigurationArgument::OverrideMaxTextureSize(256usize);
        assert_eq!(arg.to_os_string(), "-overrideMaxTextureSize 256");

        let arg = ConfigurationArgument::OverrideTextureCompression(TextureCompression::NoOverride);
        assert_eq!(arg.to_os_string(), "-overrideTextureCompression NoOverride");

        let arg = ConfigurationArgument::OverrideTextureCompression(
            TextureCompression::ForceUncompressed,
        );
        assert_eq!(
            arg.to_os_string(),
            "-overrideTextureCompression ForceUncompressed"
        );

        let arg = ConfigurationArgument::OverrideTextureCompression(
            TextureCompression::ForceFastCompressor,
        );
        assert_eq!(
            arg.to_os_string(),
            "-overrideTextureCompression ForceFastCompressor"
        );

        let arg = ConfigurationArgument::OverrideTextureCompression(
            TextureCompression::ForceNoCrunchCompression,
        );
        assert_eq!(
            arg.to_os_string(),
            "-overrideTextureCompression ForceNoCrunchCompression"
        );

        let arg = ConfigurationArgument::SilentCrashes;
        assert_eq!(arg.to_os_string(), "-silent-crashes");

        let arg = ConfigurationArgument::UpmLogFile(PathBuf::from("/var/log/unity/upm.log"));
        assert_eq!(arg.to_os_string(), "-upmLogFile /var/log/unity/upm.log");

        let arg = ConfigurationArgument::Username(OsString::from("kitsucafe"));
        assert_eq!(arg.to_os_string(), "-username kitsucafe");

        let arg = ConfigurationArgument::VcsMode(VcsMode::VisibleMetaFiles);
        assert_eq!(arg.to_os_string(), "-vcsMode \"Visible Meta Files\"");

        let arg = ConfigurationArgument::VcsMode(VcsMode::HiddenMetaFiles);
        assert_eq!(arg.to_os_string(), "-vcsMode \"Hidden Meta Files\"");

        let arg = ConfigurationArgument::VcsMode(VcsMode::PlasticScm);
        assert_eq!(arg.to_os_string(), "-vcsMode PlasticSCM");

        let arg = ConfigurationArgument::VcsMode(VcsMode::Perforce(CommandArgs(vec![
            PerforceArgument::Workspace("TestWorkspace".into()),
            PerforceArgument::Host("Hostname".into()),
            PerforceArgument::Server("8.8.8.8".into()),
            PerforceArgument::SharedLogLevel(SharedLogLevel::Info),
            PerforceArgument::Username("kitsucafe".into()),
            PerforceArgument::Password("hunter12".into()),
        ])));
        assert_eq!(
            arg.to_os_string(),
            "-vcsMode Perforce -vcPerforceWorkspace TestWorkspace -vcPerforceHost Hostname -vcPerforceServer 8.8.8.8 -vcSharedLogLevel info -vcPerforceUsername kitsucafe -vcPerforcePassword hunter12"
        );

        let arg = ConfigurationArgument::Version;
        assert_eq!(arg.to_os_string(), "-version");

        let arg = ConfigurationArgument::Timestamps;
        assert_eq!(arg.to_os_string(), "-timestamps");
    }
}
